#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "psystem.h"
#include "texture.h"

#define MIN(a, b)		((a) < (b) ? (a) : (b))

struct Ray {
	Vector3 origin, dir;
};

void display(void);
void draw_grid(float spacing, float size);
void reshape(int w, int h);
void mouse(int button, int state, int x, int y);
void motion(int x, int y);
void keyb(unsigned char key, int x, int y);
void skeyb(int key, int x, int y);
void idle(void);

void init_fire(PSystem *ps);
void init_smoke(PSystem *ps);
void init_sparks(PSystem *ps);

static Ray calc_mouse_ray(int x, int y);
static bool isect_ground(const Ray &ray, Vector3 *pt);

#define NUM_PSYS	3

PSystem psys[NUM_PSYS];
int active;

int main(int argc, char** argv) {
	glutInitWindowSize(800, 600);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutCreateWindow("Particle System");

	init_fire(psys);
	init_smoke(psys + 1);
	init_sparks(psys + 2);

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);

	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutSpecialFunc(skeyb);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutIdleFunc(idle);

	glutMainLoop();
	return 0;
}

static int prev_x = -1, prev_y = -1;
static double phi = 25;
static double theta = 0; /*spherical coordinates*/
static double distance = 10;

static int sim_interval = 33;

void display (void) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); //clear buffers
	glMatrixMode(GL_MODELVIEW);	// which matrix we want to change
	glLoadIdentity();				// model view matrix = I

	/*
		other way:
		gluLookAt(0, 0, 10, 0, 0, 0, 0, 1, 0);
		(position, target, up vector)
	 */

	glTranslatef(0, 0, -distance);		/* set the camera to (0, 0, 10) => move
									   everything 10 units along the z axis*/

	glRotatef(phi, 1, 0, 0);
	glRotatef(theta, 0, 1, 0);

	draw_grid(2.5, 50);

	psys[active].render();
	glutSwapBuffers();
}

void draw_grid(float spacing, float size)
{
	int nlines = size / spacing + 1;
	float hsz = size / 2.0;

	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);

	glBegin(GL_LINES);
	glColor3f(0.4, 0.4, 0.4);
	for(int i=0; i<nlines; i++) {
		float offs = (float)i * spacing - hsz;
		glVertex3f(-hsz, 0, offs);
		glVertex3f(hsz, 0, offs);
		glVertex3f(offs, 0, -hsz);
		glVertex3f(offs, 0, hsz);
	}
	glEnd();

	glPopAttrib();
}


void reshape (int w, int h) {
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45, (double) w/h, 1, 10000);
}

static int bn;
static int mod;

void mouse (int button, int state, int x, int y) {
	bn = button;
	prev_x = x;
	prev_y = y;

	mod = glutGetModifiers();

	if(mod & GLUT_ACTIVE_CTRL) {
		Vector3 pt;
		Ray ray = calc_mouse_ray(x, y);

		if(isect_ground(ray, &pt)) {
			for(int i=0; i<NUM_PSYS; i++) {
				psys[i].set_position(pt);
			}
		}
	}
}


void motion (int x, int y) {
	switch (bn) {
	case GLUT_LEFT_BUTTON:
		if(mod & GLUT_ACTIVE_CTRL) {
			Vector3 pt;
			Ray ray = calc_mouse_ray(x, y);

			if(isect_ground(ray, &pt)) {
				for(int i=0; i<NUM_PSYS; i++) {
					psys[i].set_position(pt);
				}
			}

		} else {
			theta += x - prev_x;
			phi += y - prev_y;

			theta = theta < 0 ? 360 + theta : theta;
			theta = theta > 360 ? theta - 360 : theta;

			phi = phi < -90 ? 90 + phi : phi;
			phi = phi > 90 ? phi - 90 : phi;
		}
		break;

	case GLUT_RIGHT_BUTTON:
		distance *= (y - prev_y) * 0.01 + 1;
		if(distance < 0.0) {
			distance = 0.0;
		}
		break;
	}

	prev_x = x;
	prev_y = y;
}

void keyb(unsigned char key, int x, int y) {
	switch(key) {
	case 27:
		exit(0);
	}
}

void skeyb(int key, int x, int y)
{
	switch(key) {
	case GLUT_KEY_LEFT:
		if(active) {
			active--;
		}
		break;

	case GLUT_KEY_RIGHT:
		if(active < NUM_PSYS - 1) {
			active++;
		}
		break;
	}
}

void idle(void) {
	static int prev_upd;
	int msec = glutGet(GLUT_ELAPSED_TIME);
	int elapsed = msec - prev_upd;

	while(elapsed >= sim_interval) {
		for(int i=0; i<NUM_PSYS; i++) {
			psys[i].update(sim_interval / 1000.0);
		}

		elapsed -= sim_interval;
		prev_upd += sim_interval;
	}


	glutPostRedisplay();
}

#define INTCOL(r, g, b) Color((double)(r) / 255.0, (double)(g) / 255.0, (double)(b) / 255.0)

void init_fire(PSystem *ps)
{
	ps->set_position(Vector3(0, 0, 0));

	ps->set_force(Vector3(0, 6, 0));
	ps->set_spawn_rate(300);
	ps->set_lifetime(1.2);
	ps->set_spawn_radius(0.1);
	ps->set_spawn_cone(M_PI / 2.0);
	ps->set_start_velocity(Vector3(0, 1.5, 0));
	ps->set_damping(0.95);
	ps->set_color(INTCOL(179, 102, 28), INTCOL(382, 85, 61));
	ps->set_alpha(0.8, 0.0);
	ps->set_size(1.1, 0.25);
	ps->set_blend_func(GL_SRC_ALPHA, GL_ONE);
	ps->set_motion_blur(false);

	ps->set_texture(load_texture("psys_rgb.ppm", "psys_alpha.ppm"));
}

void init_smoke(PSystem *ps)
{
	ps->set_position(Vector3(0, 0, 0));

	ps->set_force(Vector3(0, 0.5, 0));
	ps->set_spawn_rate(20);
	ps->set_lifetime(3.8);
	ps->set_spawn_radius(0.1);
	ps->set_spawn_cone(M_PI / 2.0);
	ps->set_start_velocity(Vector3(0, 0.1, 0));
	ps->set_damping(1);
	ps->set_color(Color(0.75, 0.75, 0.75));
	ps->set_alpha(0.4, 0.0);
	ps->set_size(0.1, 1.8);
	ps->set_blend_func(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	ps->set_motion_blur(false);

	ps->set_texture(load_texture("psys_rgb.ppm", "psys_alpha.ppm"));
}

void init_sparks(PSystem *ps)
{
	ps->set_position(Vector3(0, 0, 0));

	ps->set_force(Vector3(0, -0.8, 0));
	ps->set_spawn_rate(100);
	ps->set_lifetime(6.0);
	ps->set_spawn_radius(0.05);
	ps->set_spawn_cone(M_PI / 6.0);
	ps->set_start_velocity(Vector3(0, 3, 0));
	ps->set_damping(0.99);
	ps->set_color(INTCOL(255, 120, 51), INTCOL(175, 69, 38));
	ps->set_alpha(1.0, 0.3);
	ps->set_size(4, 6);
	ps->set_blend_func(GL_SRC_ALPHA, GL_ONE);
	ps->set_motion_blur(true);
}

static Ray calc_mouse_ray(int x, int y)
{
	Ray ray;

	double mvmat[16], proj[16];
	int viewport[4];

	glGetDoublev(GL_MODELVIEW_MATRIX, mvmat);
	glGetDoublev(GL_PROJECTION_MATRIX, proj);
	glGetIntegerv(GL_VIEWPORT, viewport);

	y = viewport[3] - y;

	double res_x, res_y, res_z;
	gluUnProject(x, y, 0.0, mvmat, proj, viewport, &res_x, &res_y, &res_z);

	ray.origin = Vector3(res_x, res_y, res_z);

	gluUnProject(x, y, 1.0, mvmat, proj, viewport, &res_x, &res_y, &res_z);
	ray.dir = Vector3(res_x, res_y, res_z) - ray.origin;

	return ray;
}

static bool isect_ground(const Ray &ray, Vector3 *pt)
{
	static const Vector3 norm(0, 1, 0);
	static const Vector3 gnd_pt(0, 0, 0);

	float ndotdir = dot(norm, ray.dir);
	if(fabs(ndotdir) < 1e-6) {
		return false;
	}

	Vector3 vovec = gnd_pt - ray.origin;
	float ndotvo = dot(norm, vovec);

	float t = ndotvo / ndotdir;

	*pt = ray.origin + ray.dir * t;
	return true;
}

