#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "matrix.h"
#include "psystem.h"
#include "texture.h"

#define EPSILON 0.00001

PSystemParameters::PSystemParameters()
{
	spawn_rate = 10.0;
	life = 5.0;
	damping = 0.995;
	spawn_radius = 0.0;
	spawn_cone = 45.0;
	start_color = end_color = Color(1, 1, 1);
	start_alpha = end_alpha = 1.0;
	start_size = end_size = 1.0;
	tex = 0;
	mblur = false;
	src_blend = GL_SRC_ALPHA;
	dst_blend = GL_ONE;
}


PSystem::PSystem() {
	pos = Vector3(0.0, 0.0, 0.0);
	time = 0;
}

void PSystem::set_position(const Vector3 &pos)
{
	this->pos = pos;
}

// particle system parameter setters
void PSystem::set_force(const Vector3 &f)
{
	param.f = f;
}

void PSystem::set_start_velocity(const Vector3 &vel)
{
	param.start_vel = vel;
}

void PSystem::set_spawn_rate(double pps)
{
	param.spawn_rate = pps;
}

void PSystem::set_lifetime(double life)
{
	param.life = life;
}

void PSystem::set_damping(double dfactor)
{
	param.damping = dfactor;
}

void PSystem::set_spawn_radius(double rad)
{
	param.spawn_radius = rad;
}

void PSystem::set_spawn_cone(double angle)
{
	param.spawn_cone = angle;
}

void PSystem::set_color(const Color &col)
{
	param.start_color = param.end_color = col;
}

void PSystem::set_color(const Color &c0, const Color &c1)
{
	param.start_color = c0;
	param.end_color = c1;
}

void PSystem::set_alpha(double alpha)
{
	param.start_alpha = param.end_alpha = alpha;
}

void PSystem::set_alpha(double a0, double a1)
{
	param.start_alpha = a0;
	param.end_alpha = a1;
}

void PSystem::set_size(double size)
{
	param.start_size = param.end_size = size;
}

void PSystem::set_size(double sz0, double sz1)
{
	param.start_size = sz0;
	param.end_size = sz1;
}

void PSystem::set_texture(unsigned int tex)
{
	param.tex = tex;
}

bool PSystem::set_texture(const char *fname)
{
	if(!(param.tex = load_texture(fname))) {
		return false;
	}
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	return true;
}

void PSystem::set_motion_blur(bool enable)
{
	param.mblur = enable;
}

void PSystem::set_blend_func(unsigned int s, unsigned int d)
{
	param.src_blend = s;
	param.dst_blend = d;
}

void PSystem::update(double dt) {
	time += dt;

	spawn();
	death();
	simulate(dt);
}

void PSystem::spawn() {
	/* number of particles to spawn */
	int spawn_count = (time - last_spawn) * param.spawn_rate;

	for (int i = 0; i < spawn_count; i++) {
		Particle *np = new Particle;

		np->set_start_position(pos, param.spawn_radius);
		np->set_start_velocity(param.start_vel, param.spawn_cone);
		np->spawn_time = time;

		particles.push_back(np);
		last_spawn = time;
	}
}

void PSystem::death() {

	std::list<Particle*>::iterator it = particles.begin();

	while (it != particles.end()) {
		std::list<Particle*>::iterator curr_it = it;
		Particle* p = *it++;

		if (time - p->spawn_time > param.life) {
			particles.erase(curr_it);
			delete p;
		}
	}
}

void PSystem::simulate(double dt) {

	std::list<Particle*>::iterator it = particles.begin();
	int i = 0;

	while (it != particles.end()) {
		i++;

		Particle* p = *it++;
		Vector3 a = param.f;		/* a = f/m but m = 1 */

		p->prev_pos = p->pos;
		p->pos = p->pos + p->v * dt;
		p->v = p->v * param.damping + a * dt;

		/* t ranges from 0 to 1 during the lifetime of the particle */
		double t = (time - p->spawn_time) / param.life;

		p->color = param.start_color + (param.end_color - param.start_color) * t;
		p->alpha = param.start_alpha + (param.end_alpha - param.start_alpha) * t;
		p->size = param.start_size + (param.end_size - param.start_size) * t;
	}
}

void PSystem::render() {
	glPushAttrib(GL_ENABLE_BIT | GL_TEXTURE_BIT | GL_COLOR_BUFFER_BIT | GL_LINE_BIT);

	glEnable(GL_BLEND);
	glBlendFunc(param.src_blend, param.dst_blend);

	glDepthMask(0);

	glLineWidth(2.0);

	if(param.tex) {
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, param.tex);
	}

	if(param.mblur) {
		glBegin(GL_LINES);
	}

	/* render all points */
	std::list<Particle*>::iterator it = particles.begin();
	while (it != particles.end()) {
		Particle* p = *it++;

		if(param.mblur) {
			Vector3 dir = p->prev_pos - p->pos;
			Vector3 tail;
			/*double len = length(dir);

			if(len > 1e-8) {
				dir = dir / len;
			}*/
			tail = p->pos + dir * p->size;

			glColor4f(p->color.x, p->color.y, p->color.z, p->alpha);
			glTexCoord2f(0, 0);
			glVertex3f(p->pos.x, p->pos.y, p->pos.z);
			glColor4f(p->color.x, p->color.y, p->color.z, 0.0);
			glTexCoord2f(1, 1);
			glVertex3f(tail.x, tail.y, tail.z);
		} else {
			billboard(p->pos, p->color, p->alpha, p->size);
		}
	}

	if(param.mblur) {
		glEnd();
	}

	glDepthMask(1);

	glPopAttrib();
}
